@@ -54,6 +54,7 @@ static const GEnumValue _pgm_drawable_property_values[] = {
   { PGM_DRAWABLE_FG_COLOR, "PGM_DRAWABLE_FG_COLOR", "drawable-fg-color" },
   { PGM_DRAWABLE_OPACITY, "PGM_DRAWABLE_OPACITY", "drawable-opacity" },
   { PGM_DRAWABLE_REGENERATE, "PGM_DRAWABLE_REGENERATE", "drawable-regenerate" },
+  { PGM_IMAGE_DATA_FILE, "PGM_IMAGE_DATA_FILE", "image-data-file" },
   { PGM_IMAGE_DATA_BUFFER, "PGM_IMAGE_DATA_BUFFER", "image-data-buffer" },
   { PGM_IMAGE_DATA_GST_BUFFER, "PGM_IMAGE_DATA_GST_BUFFER", "image-data-gst-buffer" },
   { PGM_IMAGE_DATA_PIXBUF, "PGM_IMAGE_DATA_PIXBUF", "image-data-pixbuf" },
@@ -65,6 +66,7 @@ static const GEnumValue _pgm_drawable_property_values[] = {
   { PGM_IMAGE_ALIGNMENT, "PGM_IMAGE_ALIGNMENT", "image-alignment" },
   { PGM_IMAGE_LAYOUT, "PGM_IMAGE_LAYOUT", "image-layout" },
   { PGM_IMAGE_INTERP, "PGM_IMAGE_INTERP", "image-interp" },
+  { PGM_IMAGE_WRAPPING, "PGM_IMAGE_WRAPPING", "image-wrapping" },
   { PGM_IMAGE_ASPECT_RATIO, "PGM_IMAGE_ASPECT_RATIO", "image-aspect-ratio" },
   { PGM_IMAGE_BORDER_WIDTH, "PGM_IMAGE_BORDER_WIDTH", "image-border-width" },
   { PGM_IMAGE_BORDER_INNER_COLOR, "PGM_IMAGE_BORDER_INNER_COLOR", "image-border-inner-color" },
@@ -300,6 +302,23 @@ pgm_image_interp_type_get_type (void)
   return type;
 }
 
+static const GEnumValue _pgm_image_wrapping_values[] = {
+  { PGM_IMAGE_CLAMP, "PGM_IMAGE_CLAMP", "clamp" },
+  { PGM_IMAGE_REPEAT, "PGM_IMAGE_REPEAT", "repeat" },
+  { 0, NULL, NULL }
+};
+
+GType
+pgm_image_wrapping_get_type (void)
+{
+  static GType type = 0;
+
+  if (!type)
+    type = g_enum_register_static ("PgmImageWrapping", _pgm_image_wrapping_values);
+
+  return type;
+}
+
 static const GFlagsValue _pgm_image_pixel_format_values[] = {
   { PGM_IMAGE_RGB, "PGM_IMAGE_RGB", "rgb" },
   { PGM_IMAGE_BGR, "PGM_IMAGE_BGR", "bgr" },
@@ -325,6 +344,7 @@ pgm_image_pixel_format_get_type (void)
 
 static const GEnumValue _pgm_image_storage_type_values[] = {
   { PGM_IMAGE_EMPTY, "PGM_IMAGE_EMPTY", "empty" },
+  { PGM_IMAGE_FILE, "PGM_IMAGE_FILE", "file" },
   { PGM_IMAGE_BUFFER, "PGM_IMAGE_BUFFER", "buffer" },
   { PGM_IMAGE_GST_BUFFER, "PGM_IMAGE_GST_BUFFER", "gst-buffer" },
   { PGM_IMAGE_PIXBUF, "PGM_IMAGE_PIXBUF", "pixbuf" },
