@@ -116,46 +116,42 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-#include <glib/gstdio.h> /* g_fopen */
 #include "pgmimage.h"
 #include "pgmmarshal.h"
-
-/* Number of bytes read at each callback */
-#define ASYNC_LOADER_CHUNK_SIZE 1024
+#include <stdlib.h>     /* atoi */
 
 GST_DEBUG_CATEGORY_STATIC (pgm_image_debug);
 #define GST_CAT_DEFAULT pgm_image_debug
 
 /* Image signals */
 enum {
-  PIXBUF_LOADED,
+  FILE_LOADED,
+  CLONED,
+  UN_CLONED,
   LAST_SIGNAL
 };
 
-/* Asynchronous pixbuf loader */
+/* Structure used by the thread pool to load images */
 typedef struct {
-  PgmImage        *image;    /* Image object */
-  FILE            *stream;   /* File handle */
-  GdkPixbufLoader *loader;   /* GDK Pixbuf loader */
-  GMutex          *lock;     /* Access lock: avoid holding both this and the
-                                image lock. Only exception: you must have the
-                                image lock held when calling
-                                async_loader_free().
-                                */
-  gboolean         loaded;   /* Is the loading finished? */
-  guint            max_size; /* Scaling threshold */
-  guint            source;   /* Pigment main loop source id */
-} AsyncLoader;
+  PgmImage  *image;
+  gchar     *filename;
+  GdkPixbuf *pixbuf;
+  gboolean   emit_signal;
+  guint      max_size;
+  guint      id;
+} PixbufLoaderData;
+
+/* Image internal flags */
+enum {
+  PGM_IMAGE_STORED_FROM_FILE_LOADED = (1 << 0)
+};
 
 static PgmDrawableClass *parent_class = NULL;
 static guint pgm_image_signals[LAST_SIGNAL] = { 0 };
 
-static void async_loader_free (AsyncLoader *async);
-
 /* Private methods */
 
-/* Update the ratio of the slaves.
- * Called with Image LOCK. */
+/* Update the ratio of the slaves. Called with Image LOCK. */
 static void
 update_slaves_ratio (PgmImage *image)
 {
@@ -172,8 +168,150 @@ update_slaves_ratio (PgmImage *image)
     }
 }
 
-/* Do the actual clearing of image.
+/* Cancel loading tasks in the thread pool for an image.
  * Called with Image LOCK. */
+static void
+cancel_async_loading_task (PgmImage *image)
+{
+  image->loader_id = 0;
+}
+
+/* Get whether or not a loading has been cancelled */
+static gboolean
+is_loading_cancelled (PixbufLoaderData *loader)
+{
+  PgmImage *image = loader->image;
+  guint id;
+
+  GST_OBJECT_LOCK (image);
+  id = image->loader_id;
+  GST_OBJECT_UNLOCK (image);
+
+  return (id != loader->id);
+}
+
+/* Function used by the image loader system to finish the loading in the
+ * Pigment main loop */
+static gboolean
+pixbuf_loaded_from_file (gpointer data)
+{
+  PixbufLoaderData *loader = (PixbufLoaderData*) data;
+  GdkPixbuf *pixbuf = loader->pixbuf;
+  PgmImage *image = loader->image;
+
+  /* Clear the current storage */
+  pgm_image_clear (image);
+
+  /* Setup the new storage */
+  GST_OBJECT_LOCK (image);
+  image->storage_type = PGM_IMAGE_FILE;
+  image->data.file.pixbuf = gdk_pixbuf_ref (pixbuf);
+  image->data.file.filename = loader->filename;
+  image->data.file.width = gdk_pixbuf_get_width (pixbuf);
+  image->data.file.height = gdk_pixbuf_get_height (pixbuf);
+  image->flags |= PGM_IMAGE_STORED_FROM_FILE_LOADED;
+  image->par_n = image->data.file.width;
+  image->par_d = image->data.file.height;
+  update_slaves_ratio (image);
+  GST_OBJECT_UNLOCK (image);
+
+  /* Emit signals */
+  if (loader->emit_signal)
+    g_signal_emit (G_OBJECT (image), pgm_image_signals[FILE_LOADED], 0);
+  _pgm_drawable_emit_changed (PGM_DRAWABLE (image), PGM_IMAGE_DATA_FILE);
+
+  /* Clean up */
+  g_slice_free (PixbufLoaderData, loader);
+  g_object_unref (pixbuf);
+  gst_object_unref (image);
+
+  /* Remove the idle source from the Pigment main loop */
+  return FALSE;
+}
+
+/* Function used by the thread pool to load images */
+static void
+pixbuf_loader (gpointer data,
+               gpointer user_data)
+{
+  PixbufLoaderData *loader = (PixbufLoaderData*) data;
+  PgmImage *image = loader->image;
+
+  /* Check for cancellation before loading */
+  if (is_loading_cancelled (loader))
+    {
+      GST_LOG_OBJECT (image, "\"%s\" loading has been cancelled",
+                      loader->filename);
+      goto cancel;
+    }
+
+  /* Load the image with requested scaling keeping the apect-ratio */
+  loader->pixbuf = gdk_pixbuf_new_from_file_at_size (loader->filename,
+                                                     loader->max_size,
+                                                     loader->max_size, NULL);
+
+  /* Check validity */
+  if (loader->pixbuf == NULL)
+    {
+      GST_WARNING_OBJECT (image, "can't load image from \"%s\"",
+                          loader->filename);
+      goto cancel;
+    }
+  /* Check for cancellation after loading */
+  else if (is_loading_cancelled (loader))
+    {
+      GST_LOG_OBJECT (image, "\"%s\" loading has been cancelled",
+                      loader->filename);
+      g_object_unref (loader->pixbuf);
+      goto cancel;
+    }
+  /* Finish the loading in the Pigment main loop */
+  else
+    {
+      g_idle_add (pixbuf_loaded_from_file, loader);
+      return;
+    }
+
+ cancel:
+  g_free (loader->filename);
+  g_slice_free (PixbufLoaderData, loader);
+  gst_object_unref (image);
+}
+
+/* Load an image from a file.
+ * FIXME: The static counter is not thread-safe, locking the counter
+ *        assignment and incrementation would be safer here. */
+static void
+set_from_file (PgmImage *image,
+               const gchar *filename,
+               guint max_size,
+               gboolean emit_signal)
+{
+  PixbufLoaderData *loader;
+  static guint counter = 1;
+
+  /* Create and fill loader data */
+  loader = g_slice_new (PixbufLoaderData);
+  loader->image = gst_object_ref (image);
+  loader->filename = g_strdup (filename);
+  loader->max_size = (max_size < 1) ? -1 : max_size;
+  loader->emit_signal = emit_signal;
+  GST_OBJECT_LOCK (image);
+  image->loader_id = counter;
+  GST_OBJECT_UNLOCK (image);
+  loader->id = counter;
+
+  /* Increment counter for the next loader, avoid zero since it's used by
+   * images to cancel loading tasks */
+  if (G_UNLIKELY (++counter == 0))
+    counter++;
+
+  /* Push the loading in the thread pool */
+  g_thread_pool_push (PGM_IMAGE_GET_CLASS (image)->pixbuf_loader_pool,
+                      loader, NULL);
+}
+
+/* Do the actual clearing of image. Called with Image LOCK. */
 static PgmError
 do_clear (PgmImage *image)
 {
@@ -181,19 +319,34 @@ do_clear (PgmImage *image)
 
   switch (image->storage_type)
     {
+    case PGM_IMAGE_FILE:
+      GST_DEBUG_OBJECT (image, "Clearing file image");
+      if (image->flags & PGM_IMAGE_STORED_FROM_FILE_LOADED)
+        {
+          gdk_pixbuf_unref ((GdkPixbuf*) (image->data.file.pixbuf));
+          image->data.file.pixbuf = NULL;
+          g_free (image->data.file.filename);
+          image->data.file.filename = NULL;
+          image->flags &= ~PGM_IMAGE_STORED_FROM_FILE_LOADED;
+        }
+      break;
+
     case PGM_IMAGE_BUFFER:
       GST_DEBUG_OBJECT (image, "Clearing buffer image");
       g_free (image->data.buffer.buffer);
+      image->data.buffer.buffer = NULL;
       break;
 
     case PGM_IMAGE_GST_BUFFER:
       GST_DEBUG_OBJECT (image, "Clearing GstBuffer image");
       gst_buffer_unref (image->data.gst_buffer.gst_buffer);
+      image->data.gst_buffer.gst_buffer = NULL;
       break;
 
     case PGM_IMAGE_PIXBUF:
       GST_DEBUG_OBJECT (image, "Clearing GdkPixbuf image");
-      gdk_pixbuf_unref ((GdkPixbuf *) (image->data.pixbuf.pixbuf));
+      gdk_pixbuf_unref ((GdkPixbuf*) (image->data.pixbuf.pixbuf));
+      image->data.pixbuf.pixbuf = NULL;
       break;
 
     case PGM_IMAGE_SYSTEM_BUFFER:
@@ -205,11 +358,16 @@ do_clear (PgmImage *image)
       GST_DEBUG_OBJECT (image, "Clearing slaved image");
       if (image->master)
         {
+          PgmImage *master;
+
           /* Remove ourself from the master's slave list. */
           GST_OBJECT_LOCK (image->master);
           image->master->slaves = g_list_remove (image->master->slaves, image);
+          master = image->master;
           GST_OBJECT_UNLOCK (image->master);
           image->master = NULL;
+          g_signal_emit (G_OBJECT (master), pgm_image_signals[UN_CLONED], 0,
+                         g_list_length (master->slaves));
         }
       break;
 
@@ -225,210 +383,8 @@ do_clear (PgmImage *image)
   return PGM_ERROR_OK;
 }
 
-/* Asynchronous loader chunk reader callback */
-static gint
-async_loader_read_chunk_cb (gpointer data)
-{
-  AsyncLoader *async = (AsyncLoader*) data;
-  PgmImage *image = async->image;
-  gchar buffer[ASYNC_LOADER_CHUNK_SIZE];
-  gsize bytes_read;
-  GdkPixbuf *pixbuf;
-
-  /* we want to avoid a disposal of the image between here and when we use it */
-  gst_object_ref (image);
-
-  g_mutex_lock (async->lock);
-
-  /* the source may have been destroyed while we were waiting for the lock (e.g.
-   * by async_loader_free() which holds that lock) */
-  if (g_source_is_destroyed (g_main_current_source ()))
-    {
-      if (async->lock)
-        g_mutex_unlock (async->lock);
-      gst_object_unref (image);
-      return FALSE;
-    }
-
-  /* Read the chunk */
-  bytes_read = fread (buffer, 1, ASYNC_LOADER_CHUNK_SIZE, async->stream);
-  if (ferror (async->stream))
-    {
-      GST_WARNING_OBJECT (image, "cannot read from the file: %s",
-                        g_strerror (errno));
-      goto error;
-    }
-
-  /* Write the chunk */
-  if (!gdk_pixbuf_loader_write (async->loader, (const guchar*) buffer,
-                                bytes_read, NULL))
-    {
-      GST_WARNING_OBJECT (image, "GdkPixbufLoader cannot parse the buffer");
-      goto error;
-    }
-
-  /* Image data not fully loaded */
-  if (!feof (async->stream))
-    {
-      g_mutex_unlock (async->lock);
-      gst_object_unref (image);
-      return TRUE;
-    }
-  /* Image data fully loaded */
-  else
-    {
-      /* Closing makes the loader aware that its got the full file: without this
-       * it's buggy with small files */
-      gdk_pixbuf_loader_close (async->loader, NULL);
-
-      /* Get the pixbuf */
-      pixbuf = gdk_pixbuf_ref (gdk_pixbuf_loader_get_pixbuf (async->loader));
-      if (!pixbuf)
-        {
-          GST_WARNING_OBJECT (image, "GdkPixbufLoader cannot get the pixbuf");
-          goto error;
-        }
-
-      /* Now we don't need to access the async loader anymore, but we want to
-       * get a lock on the image, which we should not do with the async loader
-       * lock held */
-      g_mutex_unlock (async->lock);
-
-      /* Clean up the previous data */
-      pgm_image_clear (image);
-
-      GST_OBJECT_LOCK (image);
-      /* Fill the new storage informations */
-      image->storage_type = PGM_IMAGE_PIXBUF;
-      image->par_n = gdk_pixbuf_get_width (pixbuf);
-      image->par_d = gdk_pixbuf_get_height (pixbuf);
-      image->data.pixbuf.pixbuf = pixbuf;
-      update_slaves_ratio (image);
-      GST_OBJECT_UNLOCK (image);
-
-
-      /* And emit signals */
-      _pgm_drawable_emit_changed (PGM_DRAWABLE (image), PGM_IMAGE_DATA_PIXBUF);
-      g_signal_emit (G_OBJECT (image), pgm_image_signals[PIXBUF_LOADED], 0);
-
-      /* We don't need the loader anymore */
-      GST_OBJECT_LOCK (image);
-      /* we free image->async_loader and not async since it may have been freed
-       * and NULL-ified since we unlocked it */
-      async_loader_free (image->async_loader);
-      image->async_loader = NULL;
-      GST_OBJECT_UNLOCK (image);
-
-      gst_object_unref (image);
-      return FALSE;
-    }
-
- error:
-  fclose (async->stream);
-  async->stream = NULL;
-  gdk_pixbuf_loader_close (async->loader, NULL);
-  g_object_unref (async->loader);
-  async->loader = NULL;
-  async->loaded = TRUE;
-  g_mutex_unlock (async->lock);
-  gst_object_unref (image);
-  return FALSE;
-}
-
-/* GdkPixbufLoader handler called whenever loader's figured out the size */
-static void
-async_loader_size_prepared_cb (GdkPixbufLoader *loader,
-                               gint width,
-                               gint height,
-                               gpointer data)
-{
-  AsyncLoader *async = (AsyncLoader*) data;
-  gint max_size = (gint) async->max_size;
-
-  /* Request a scaling if the width or the height exceeds the maximum size */
-  if (width > max_size || height > max_size)
-    {
-      if (width >= height)
-        {
-          height = height * max_size / width;
-          width = max_size;
-        }
-      else
-        {
-          width = width * max_size / height;
-          height = max_size;
-        }
-
-      gdk_pixbuf_loader_set_size (loader, width, height);
-    }
-}
-
-/* Asynchronous loader creation */
-static AsyncLoader*
-async_loader_new (PgmImage *image,
-                  FILE *stream,
-                  guint max_size)
-{
-  AsyncLoader *async;
-
-  async = g_slice_new0 (AsyncLoader);
-  if (G_UNLIKELY (!async))
-    {
-      GST_WARNING_OBJECT (image, "cannot create the asynchronous image loader");
-      return NULL;
-    }
-
-  async->image = image;
-  async->stream = stream;
-  async->max_size = max_size;
-  async->loaded = FALSE;
-  async->lock = g_mutex_new ();
-  async->loader = gdk_pixbuf_loader_new ();
-
-  /* Connect to 'size_prepared' signal to adapt image size as requested */
-  if (max_size != 0)
-    g_signal_connect (async->loader, "size_prepared",
-                      G_CALLBACK (async_loader_size_prepared_cb), async);
-
-  /* Add the progressive loading source to the main Pigment loop */
-  async->source = g_idle_add (async_loader_read_chunk_cb, async);
-
-  return async;
-}
-
-/* Asynchronous loader destruction.
- * Note: This function should be called with the lock on the corresponding image
- * held.
- */
-static void
-async_loader_free (AsyncLoader *async)
-{
-  if (async == NULL)
-    return;
-
-  if (!async->loaded)
-    {
-      g_source_remove (async->source);
-
-      /* from this point on, async_loader_read_chunk_cb() cannot be called, we
-       * lock in case it has been called before the source removal. */
-      g_mutex_lock (async->lock);
-      g_mutex_unlock (async->lock);
-      fclose (async->stream);
-      async->stream = NULL;
-      gdk_pixbuf_loader_close (async->loader, NULL);
-      g_object_unref (async->loader);
-      async->loader = NULL;
-    }
-
-  g_mutex_free (async->lock);
-  async->lock = NULL;
-
-  g_slice_free (AsyncLoader, async);
-}
-
-/* Gets the width and height in pixel of the original image buffer.
- * Called with LOCK. */
+/* Gets the width and height in pixel of the original image buffer. Called
+ * with LOCK. */
 static void
 image_get_buffer_pixel_size (PgmImage *image,
                              guint *width,
@@ -436,6 +392,11 @@ image_get_buffer_pixel_size (PgmImage *image,
 {
   switch (image->storage_type)
     {
+    case PGM_IMAGE_FILE:
+      *width = image->data.file.width;
+      *height = image->data.file.height;
+      break;
+
     case PGM_IMAGE_BUFFER:
       *width = image->data.buffer.width;
       *height = image->data.buffer.height;
@@ -467,8 +428,7 @@ image_get_buffer_pixel_size (PgmImage *image,
 
 /* Gets the position of the upper left corner of the image in drawable
  * coordinates (in x_offset and y_offset), and the resolution of the image, in
- * pixel per canvas unit.
- * Called with LOCK. */
+ * pixel per canvas unit. Called with LOCK. */
 static void
 image_get_offset_and_resolution (PgmImage *image,
                                  gfloat *x_offset,
@@ -497,7 +457,7 @@ image_get_offset_and_resolution (PgmImage *image,
   fill_height = (image->layout == PGM_IMAGE_SCALED && !image_wider)
                 || (image->layout == PGM_IMAGE_ZOOMED && image_wider);
 
-  /* resolution */
+  /* Resolution */
   if (fill_height)
     {
       *x_resolution = image_width / (image_aspect_ratio * drawable->height);
@@ -528,18 +488,47 @@ image_get_offset_and_resolution (PgmImage *image,
 
 /* GObject stuff */
 
-G_DEFINE_TYPE (PgmImage, pgm_image, PGM_TYPE_DRAWABLE);
+/* Forward declarations */
+static void pgm_image_base_class_init (gpointer g_class);
+static void pgm_image_base_class_finalize (gpointer g_class);
+static void pgm_image_class_init (PgmImageClass *klass);
+static void pgm_image_init (PgmImage *image);
+
+GType
+pgm_image_get_type (void)
+{
+  static GType image_type = 0;
+
+  if (G_UNLIKELY (image_type == 0))
+    {
+      static const GTypeInfo image_info = {
+        sizeof (GstElementClass),
+        (GBaseInitFunc) pgm_image_base_class_init,
+        (GBaseFinalizeFunc) pgm_image_base_class_finalize,
+        (GClassInitFunc) pgm_image_class_init,
+        NULL,
+        NULL,
+        sizeof (PgmImage),
+        0,
+        (GInstanceInitFunc) pgm_image_init,
+        NULL
+      };
+
+      image_type = g_type_register_static (PGM_TYPE_DRAWABLE, "PgmImage",
+                                           &image_info, 0);
+    }
+
+  GST_DEBUG_CATEGORY_INIT (pgm_image_debug, "pgm_image", 0,
+                           "Pigment Image object");
+
+  return image_type;
+}
 
 static void
 pgm_image_dispose (GObject *object)
 {
   PgmImage *image = PGM_IMAGE (object);
 
-  GST_OBJECT_LOCK (image);
-  async_loader_free (image->async_loader);
-  image->async_loader = NULL;
-  GST_OBJECT_UNLOCK (image);
-
   /* Image is a master */
   if (image->slaves)
     {
@@ -568,6 +557,35 @@ pgm_image_dispose (GObject *object)
   GST_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
 }
 
+static void
+pgm_image_base_class_init (gpointer g_class)
+{
+  PgmImageClass *klass = PGM_IMAGE_CLASS (g_class);
+  const gchar *env_var = NULL;
+  gint max_threads = 1;
+
+  /* Get the number of threads to use in the pool from the environment. The
+   * number is 1 by default. This is there to ease us to make a decision on
+   * the default number to use depending on the CPU technology (multi-core,
+   * hyper-threading, ...). */
+  env_var = g_getenv ("PGM_IMAGE_THREAD_POOL_SIZE");
+  if (env_var)
+    max_threads = CLAMP (atoi (env_var), 1, 16);
+  klass->pixbuf_loader_pool =
+    g_thread_pool_new (pixbuf_loader, NULL, max_threads, TRUE, NULL);
+
+  GST_DEBUG ("Image thread pool size: %d", max_threads);
+}
+
+static void
+pgm_image_base_class_finalize (gpointer g_class)
+{
+  PgmImageClass *klass = PGM_IMAGE_CLASS (g_class);
+
+  g_thread_pool_free (klass->pixbuf_loader_pool, TRUE, TRUE);
+  klass->pixbuf_loader_pool = NULL;
+}
+
 static void
 pgm_image_class_init (PgmImageClass *klass)
 {
@@ -576,22 +594,47 @@ pgm_image_class_init (PgmImageClass *klass)
   parent_class = g_type_class_peek_parent (klass);
 
   /**
-   * PgmImage::pixbuf-loaded:
+   * PgmImage::file-loaded:
    * @image: the #PgmImage
    *
    * Will be emitted after @image has finished to load its data from the file
    * path given in the pgm_image_set_from_file() method.
    */
-  pgm_image_signals[PIXBUF_LOADED] =
-    g_signal_new ("pixbuf-loaded", G_TYPE_FROM_CLASS (klass),
+  pgm_image_signals[FILE_LOADED] =
+    g_signal_new ("file-loaded", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST,
-                  G_STRUCT_OFFSET (PgmImageClass, pixbuf_loaded),
+                  G_STRUCT_OFFSET (PgmImageClass, file_loaded),
                   NULL, NULL, pgm_marshal_VOID__VOID, G_TYPE_NONE, 0);
+  /**
+   * PgmImage::cloned:
+   * @image: the #PgmImage
+   * @clone: the cloned #PgmImage
+   * @n_clones: the number of cloned images
+   *
+   * Will be emitted after @image has been cloned by another #PgmImage through
+   * the pgm_image_new_from_image() or pgm_image_set_from_image() functions.
+   */
+  pgm_image_signals[CLONED] =
+    g_signal_new ("cloned", G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  G_STRUCT_OFFSET (PgmImageClass, cloned),
+                  NULL, NULL, pgm_marshal_VOID__OBJECT_UINT, G_TYPE_NONE, 2,
+                  PGM_TYPE_IMAGE, G_TYPE_UINT);
+  /**
+   * PgmImage::un-cloned:
+   * @image: the #PgmImage
+   * @n_clones: the number of cloned images
+   *
+   * Will be emitted after @image has got a cloned #PgmImage removed.
+   */
+  pgm_image_signals[UN_CLONED] =
+    g_signal_new ("un-cloned", G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  G_STRUCT_OFFSET (PgmImageClass, un_cloned),
+                  NULL, NULL, pgm_marshal_VOID__UINT, G_TYPE_NONE, 1,
+                  G_TYPE_UINT);
 
   gobject_class->dispose = GST_DEBUG_FUNCPTR (pgm_image_dispose);
-
-  GST_DEBUG_CATEGORY_INIT (pgm_image_debug, "pgm_image", 0,
-                           "Pigment Image object");
 }
 
 static void
@@ -604,6 +647,8 @@ pgm_image_init (PgmImage *image)
   image->align = PGM_IMAGE_CENTER;
   image->layout = PGM_IMAGE_SCALED;
   image->interp = PGM_IMAGE_BILINEAR;
+  image->wrap_s = PGM_IMAGE_REPEAT;
+  image->wrap_t = PGM_IMAGE_REPEAT;
   image->master = NULL;
   image->slaves = NULL;
   image->border_width = 0.0f;
@@ -615,7 +660,7 @@ pgm_image_init (PgmImage *image)
   image->border_outer_g = 255;
   image->border_outer_b = 255;
   image->border_outer_a = 255;
-  image->async_loader = NULL;
+  image->flags = 0;
 }
 
 /* public methods */
@@ -646,9 +691,16 @@ pgm_image_new (void)
  * @max_size: the maximum size of the image in pixels before loading it in the
  * #PgmImage or 0 to not constrain the size.
  *
- * Creates a new #PgmImage instance loading progressively an image from the
- * given @filename. It optionally pre-scales the image so that it has a maximum
- * width and height of @max_size.
+ * Creates a new #PgmImage instance loading an image from the given @filename.
+ * It optionally pre-scales the image so that it has a maximum width and height
+ * of @max_size.
+ *
+ * The loading is done asynchronously, it means that this function returns right
+ * after the image object is created, the loading and decoding of the image
+ * being done in another thread. The <link linkend="PgmImage-file-loaded">
+ * 'file-loaded'</link> signal is emitted from the Pigment main loop (the
+ * thread running pgm_main() or the thread running a #GMainLoop using the
+ * default #GMainContext) once the loading finishes successfully.
  *
  * MT safe.
  *
@@ -674,6 +726,35 @@ pgm_image_new_from_file (const gchar *filename,
   return PGM_DRAWABLE (image);
 }
 
+/**
+ * pgm_image_new_from_pixbuf:
+ * @pixbuf: the #GdkPixbuf.
+ *
+ * Creates a new #PgmImage instance using @pixbuf as stored image.
+ *
+ * MT safe.
+ *
+ * Returns: a new #PgmImage instance.
+ */
+PgmDrawable *
+pgm_image_new_from_pixbuf (GdkPixbuf *pixbuf)
+{
+  PgmImage *image;
+  PgmError ret = PGM_ERROR_OK;
+
+  image = g_object_new (PGM_TYPE_IMAGE, NULL);
+  GST_DEBUG_OBJECT (image, "created new image from pixbuf");
+
+  ret = pgm_image_set_from_pixbuf (image, pixbuf);
+  if (PGM_ERROR_OK != ret)
+    {
+      gst_object_unref (image);
+      return NULL;
+    }
+
+  return PGM_DRAWABLE (image);
+}
+
 /**
  * pgm_image_new_from_buffer:
  * @format: the pixel format of the buffer.
@@ -741,12 +822,12 @@ pgm_image_new_from_image (PgmImage *src_image)
  * Loads an image from the file @filename. It optionally pre-scales the image so
  * it has a maximum width and height of @max_size.
  *
- * This function is meant to be asynchronous, it loads the image by small chunks
- * of 1024 bytes using an idle source in the Pigment mainloop. The consequence
- * being that the storage type of @image doesn't change immediately, but only
- * once the whole image is loaded. You can connect a callback to the
- * <link linkend="PgmImage-pixbuf-loaded">pixbuf-loaded</link> signal to know when
- * the loading is done.
+ * The loading is done asynchronously, it means that this function returns right
+ * after the image object is created, the loading and decoding of the image
+ * being done in another thread. The <link linkend="PgmImage-pixbuf-loaded">
+ * 'pixbuf-loaded'</link> signal is emitted from the Pigment main loop (the
+ * thread running pgm_main() or the thread running a #GMainLoop using the
+ * default #GMainContext) once the loading finishes successfully.
  *
  * MT safe.
  *
@@ -757,25 +838,48 @@ pgm_image_set_from_file (PgmImage *image,
                          const gchar *filename,
                          guint max_size)
 {
-  FILE *stream;
-
   g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
 
-  /* Get a file handle from the filename */
-  stream = g_fopen (filename, "rb");
-  if (!stream)
-    {
-      GST_WARNING_OBJECT (image, "cannot open %s: %s", filename,
-                        g_strerror (errno));
-      return PGM_ERROR_X;
-    }
+  if (filename == NULL)
+    return PGM_ERROR_X;
+
+  set_from_file (image, filename, max_size, TRUE);
+
+  return PGM_ERROR_OK;
+}
+
+/**
+ * pgm_image_set_from_pixbuf:
+ * @image: a #PgmImage object.
+ * @pixbuf: the #GdkPixbuf.
+
+ * Loads an image in @image from a pixbuf. @pixbuf will have its reference
+ * count increased by 1 and will not get freed until the drawable gets cleaned
+ * up or that a new buffer is loaded.
+ *
+ * MT safe.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+pgm_image_set_from_pixbuf (PgmImage *image,
+                           GdkPixbuf *pixbuf)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+  g_return_val_if_fail (pixbuf != NULL, PGM_ERROR_X);
+
+  pgm_image_clear (image);
 
-  /* Free the current asynchronous loader and create the new one */
   GST_OBJECT_LOCK (image);
-  async_loader_free ((AsyncLoader*) image->async_loader);
-  image->async_loader = (gpointer) async_loader_new (image, stream, max_size);
+  image->storage_type = PGM_IMAGE_PIXBUF;
+  image->data.pixbuf.pixbuf = gdk_pixbuf_ref (pixbuf);
+  image->par_n = gdk_pixbuf_get_width (pixbuf);
+  image->par_d = gdk_pixbuf_get_height (pixbuf);
+  update_slaves_ratio (image);
   GST_OBJECT_UNLOCK (image);
 
+  _pgm_drawable_emit_changed (PGM_DRAWABLE (image), PGM_IMAGE_DATA_PIXBUF);
+
   return PGM_ERROR_OK;
 }
 
@@ -814,6 +918,8 @@ pgm_image_set_from_buffer (PgmImage *image,
   /* Let's set the storage data */
   GST_OBJECT_LOCK (image);
 
+  cancel_async_loading_task (image);
+
   /* The buffer sent is not the first of this size, it's not needed to call
    * a clear, just free the buffer. */
   if (G_LIKELY (image->storage_type == PGM_IMAGE_BUFFER
@@ -895,6 +1001,8 @@ pgm_image_set_from_gst_buffer (PgmImage *image,
 
   GST_OBJECT_LOCK (image);
 
+  cancel_async_loading_task (image);
+
   /* The GstBuffer sent is not the first, it's not needed to call a clear,
    * just unref the GstBuffer. I mean that the implementation should not be
    * signaled to clear its allocated size/format for the stream of buffers
@@ -974,6 +1082,8 @@ pgm_image_set_from_system_buffer (PgmImage *image,
   /* Let's set the storage data */
   GST_OBJECT_LOCK (image);
 
+  cancel_async_loading_task (image);
+
   /* Let's clear if needed */
   if (image->storage_type != PGM_IMAGE_EMPTY)
     {
@@ -1038,6 +1148,7 @@ pgm_image_set_from_image (PgmImage *image,
       goto beach;
     }
 
+  cancel_async_loading_task (image);
   do_clear (image);
   GST_OBJECT_UNLOCK (image);
   _pgm_drawable_emit_changed (PGM_DRAWABLE (image), PGM_IMAGE_DATA_EMPTY);
@@ -1078,10 +1189,69 @@ pgm_image_set_from_image (PgmImage *image,
 
   _pgm_drawable_emit_changed (PGM_DRAWABLE (image), PGM_IMAGE_DATA_IMAGE);
 
+  g_signal_emit (G_OBJECT (image->master), pgm_image_signals[CLONED], 0,
+                 image, g_list_length (image->master->slaves));
+
 beach:
   return ret;
 }
 
+/**
+ * pgm_image_to_pixbuf:
+ * @image: a #PgmImage object.
+ * @pixbuf: a double pointer to #GdkPixbuf. Unref after usage.
+ *
+ * Retrieves a #GdkPixbuf of the image currently stored in @image.
+ *
+ * Only the images stored as pixbuf can be retrieved at the moment, the
+ * conversion for the other storage types is not implemented yet.
+ *
+ * MT safe.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+pgm_image_to_pixbuf (PgmImage *image,
+                     GdkPixbuf **pixbuf)
+{
+  PgmError ret = PGM_ERROR_X;
+
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+  g_return_val_if_fail (image != NULL, PGM_ERROR_X);
+
+  GST_OBJECT_LOCK (image);
+
+  switch (image->storage_type)
+    {
+    case PGM_IMAGE_PIXBUF:
+      *pixbuf = gdk_pixbuf_ref (image->data.pixbuf.pixbuf);
+      ret = PGM_ERROR_OK;
+      break;
+
+      /* FIXME: Not implemented yet */
+    case PGM_IMAGE_FILE:
+    case PGM_IMAGE_IMAGE:
+    case PGM_IMAGE_BUFFER:
+    case PGM_IMAGE_GST_BUFFER:
+    case PGM_IMAGE_SYSTEM_BUFFER:
+      GST_WARNING_OBJECT (image, "Conversion to pixbuf not implemented for raw "
+                          "buffers, GStreamer buffers, system buffers and "
+                          "cloned images");
+      break;
+
+    case PGM_IMAGE_EMPTY:
+      GST_WARNING_OBJECT (image, "Cannot convert empty image to pixbuf");
+      break;
+
+    default:
+      break;
+    }
+
+  GST_OBJECT_UNLOCK (image);
+
+  return ret;
+}
+
 /**
  * pgm_image_clear:
  * @image: a #PgmImage object.
@@ -1101,9 +1271,10 @@ pgm_image_clear (PgmImage *image)
 {
   PgmError ret = PGM_ERROR_OK;
 
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+
   GST_OBJECT_LOCK (image);
-  async_loader_free (image->async_loader);
-  image->async_loader = NULL;
+  cancel_async_loading_task (image);
   ret = do_clear (image);
   GST_OBJECT_UNLOCK (image);
 
@@ -1427,6 +1598,73 @@ pgm_image_get_interp (PgmImage *image,
   return PGM_ERROR_OK;
 }
 
+/**
+ * pgm_image_set_wrapping:
+ * @image: a #PgmImage object.
+ * @wrap_s: the wrapping on the s coordinate. #PGM_IMAGE_REPEAT by default.
+ * @wrap_t: the wrapping on the t coordinate. #PGM_IMAGE_REPEAT by default.
+ *
+ * Sets the horizontal and vertical wrapping modes used for the mapping of
+ * @image. When the mapping matrix is modified, pixels outside the range of the
+ * image content are accessed, the wrapping mode defines the behaviour of the
+ * mapping in that particular case.
+ *
+ * MT safe.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+pgm_image_set_wrapping (PgmImage *image,
+                        PgmImageWrapping wrap_s,
+                        PgmImageWrapping wrap_t)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+
+  GST_OBJECT_LOCK (image);
+
+  image->wrap_s = wrap_s;
+  image->wrap_t = wrap_t;
+
+  GST_OBJECT_UNLOCK (image);
+
+  _pgm_drawable_emit_changed (PGM_DRAWABLE (image), PGM_IMAGE_WRAPPING);
+
+  return  PGM_ERROR_OK;
+}
+
+/**
+ * pgm_image_get_wrapping:
+ * @image: a #PgmImage object.
+ * @wrap_s: a pointer to a #PgmImageInterpType where the wrapping mode on the s
+ * coordinate is going to be stored.
+ * @wrap_t: a pointer to a #PgmImageInterpType where the wrapping mode on the t
+ * coordinate is going to be stored.
+ *
+ * Retrieves in @wrap_s and @wrap_t the current wrapping modes of @image.
+ *
+ * MT safe.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+pgm_image_get_wrapping (PgmImage *image,
+                        PgmImageWrapping *wrap_s,
+                        PgmImageWrapping *wrap_t)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+  g_return_val_if_fail (wrap_s != NULL, PGM_ERROR_X);
+  g_return_val_if_fail (wrap_t != NULL, PGM_ERROR_X);
+
+  GST_OBJECT_LOCK (image);
+
+  *wrap_s = image->wrap_s;
+  *wrap_t = image->wrap_t;
+
+  GST_OBJECT_UNLOCK (image);
+
+  return PGM_ERROR_OK;
+}
+
 /**
  * pgm_image_set_aspect_ratio:
  * @image: a #PgmImage object.
@@ -1804,3 +2042,71 @@ pgm_image_to_drawable (PgmImage *image,
 
   return PGM_ERROR_OK;
 }
+
+/* Protected functions */
+
+/**
+ * _pgm_image_stored_from_file_free:
+ * @image: a #PgmImage instance
+ *
+ * Free the image buffer allocated by a pgm_image_set_from_file(). Can be used
+ * by a plugin to clean up the central memory after the image buffer has been
+ * sent to the video memory.
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+_pgm_image_stored_from_file_free (PgmImage *image)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+
+  GST_OBJECT_LOCK (image);
+
+  /* Free the pixbuf if current storage has been loaded from a file */
+  if (image->storage_type == PGM_IMAGE_FILE
+      && image->flags & PGM_IMAGE_STORED_FROM_FILE_LOADED)
+    {
+      gdk_pixbuf_unref ((GdkPixbuf*) (image->data.file.pixbuf));
+      image->data.file.pixbuf = NULL;
+      image->flags &= ~PGM_IMAGE_STORED_FROM_FILE_LOADED;
+    }
+
+  GST_OBJECT_UNLOCK (image);
+
+  return PGM_ERROR_OK;
+}
+
+/**
+ * _pgm_image_stored_from_file_load:
+ * @image: a #PgmImage instance
+ *
+ * Reload the image buffer set through pgm_image_set_from_file().
+ *
+ * Returns: a #PgmError indicating success/failure.
+ */
+PgmError
+_pgm_image_stored_from_file_load (PgmImage *image)
+{
+  g_return_val_if_fail (PGM_IS_IMAGE (image), PGM_ERROR_X);
+
+  GST_OBJECT_LOCK (image);
+
+  if (image->storage_type != PGM_IMAGE_FILE)
+    {
+      GST_OBJECT_UNLOCK (image);
+      return PGM_ERROR_X;
+    }
+
+  if (! (image->flags & PGM_IMAGE_STORED_FROM_FILE_LOADED))
+    {
+      GST_OBJECT_UNLOCK (image);
+      set_from_file (image, image->data.file.filename,
+                     MAX (image->data.file.width, image->data.file.height),
+                     FALSE);
+      return PGM_ERROR_OK;
+    }
+
+  GST_OBJECT_UNLOCK (image);
+
+  return PGM_ERROR_OK;
+}
