@@ -49,6 +49,7 @@ typedef struct _PgmImage      PgmImage;
 typedef struct _PgmImageClass PgmImageClass;
 
 /* Data storage */
+typedef struct _PgmImageFile         PgmImageFile;
 typedef struct _PgmImageBuffer       PgmImageBuffer;
 typedef struct _PgmImageGstBuffer    PgmImageGstBuffer;
 typedef struct _PgmImageSystemBuffer PgmImageSystemBuffer;
@@ -113,10 +114,10 @@ typedef enum {
 
 /**
  * PgmImageInterpType:
- * @PGM_IMAGE_NEAREST: The nearest neighbour sampling. This is the fastest
- * and lowest quality mode.
- * @PGM_IMAGE_BILINEAR: The bilinear sampling. This is the slowest and
- * highest quality mode.
+ * @PGM_IMAGE_NEAREST: The nearest neighbour sampling is the fastest and
+ * lowest quality mode.
+ * @PGM_IMAGE_BILINEAR: The bilinear sampling is the slowest and highest
+ * quality mode.
  *
  * The different image interpolation types.
  */
@@ -125,6 +126,20 @@ typedef enum {
   PGM_IMAGE_BILINEAR
 } PgmImageInterpType;
 
+/**
+ * PgmImageWrapping:
+ * @PGM_IMAGE_CLAMP: The image content is displayed in the mapping range
+ * [0.0, 1.0] and discarded outside it.
+ * @PGM_IMAGE_REPEAT: The image content is displayed in the mapping range
+ * [0.0, 1.0] and repeated infinitely outside it.
+ *
+ * The different image wrapping types.
+ */
+typedef enum {
+  PGM_IMAGE_CLAMP,
+  PGM_IMAGE_REPEAT
+} PgmImageWrapping;
+
 /**
  * PgmImagePixelFormat:
  * @PGM_IMAGE_RGB: 24 bits RGB (3 bytes, red 8 @ 16, green 8 @ 8, blue 8 @ 0).
@@ -155,6 +170,22 @@ typedef enum {
   PGM_IMAGE_YUYV = (1 << 7)
 } PgmImagePixelFormat;
 
+/**
+ * PgmImageFile:
+ * @pixbuf: the #GdkPixbuf object.
+ * @filename: the file name.
+ * @width: the cached pixbuf width.
+ * @height: the cached pixbuf height.
+ *
+ * The #PgmImageFile structure.
+ */
+struct _PgmImageFile {
+  GdkPixbuf *pixbuf;
+  gchar     *filename;
+  guint      width;
+  guint      height;
+};
+
 /**
  * PgmImageBuffer:
  * @buffer: the buffer pointer.
@@ -222,6 +253,7 @@ struct _PgmImageSystemBuffer {
 /**
  * PgmImageStorageType:
  * @PGM_IMAGE_EMPTY: no image stored.
+ * @PGM_IMAGE_FILE: a #PgmImageFile is stored.
  * @PGM_IMAGE_BUFFER: a #PgmImageBuffer is stored.
  * @PGM_IMAGE_GST_BUFFER: a #PgmImageGstBuffer is stored.
  * @PGM_IMAGE_PIXBUF: a #PgmImagePixbuf is stored.
@@ -232,6 +264,7 @@ struct _PgmImageSystemBuffer {
  */
 typedef enum {
   PGM_IMAGE_EMPTY,
+  PGM_IMAGE_FILE,
   PGM_IMAGE_BUFFER,
   PGM_IMAGE_GST_BUFFER,
   PGM_IMAGE_PIXBUF,
@@ -241,6 +274,7 @@ typedef enum {
 
 /**
  * PgmImageData:
+ * @buffer: the #PgmImageFile structure.
  * @buffer: the #PgmImageBuffer structure.
  * @gst_buffer: the #PgmImageGstBuffer structure.
  * @pixbuf: the #PgmImagePixbuf structure.
@@ -249,6 +283,7 @@ typedef enum {
  * The stored data depending on the storage type.
  */
 union _PgmImageData {
+  PgmImageFile         file;
   PgmImageBuffer       buffer;
   PgmImageGstBuffer    gst_buffer;
   PgmImagePixbuf       pixbuf;
@@ -263,6 +298,8 @@ union _PgmImageData {
  * @layout: the image layout.
  * @align: the image alignment.
  * @interp: the image interpolation.
+ * @wrap_s: the image s wrapping.
+ * @wrap_t: the image t wrapping.
  * @par_n: the image pixel-aspect-ratio numerator.
  * @par_d: the image pixel-aspect-ratio denominator.
  * @border_width: the image border width.
@@ -293,6 +330,8 @@ struct _PgmImage {
   PgmImageLayoutType  layout;
   PgmImageAlignment   align;
   PgmImageInterpType  interp;
+  PgmImageWrapping    wrap_s;
+  PgmImageWrapping    wrap_t;
   guint               par_n, par_d;
 
   /* Image border */
@@ -311,8 +350,11 @@ struct _PgmImage {
   GList    *slaves;
 
   /*< private >*/
-  /* Asynchronous image loader */
-  gpointer async_loader;
+  /* Current loader id */
+  guint loader_id;
+
+  /* Management bit mask */
+  guint8 flags;
 
   /* Padding */
   gpointer _pgm_reserved[PGM_PADDING];
@@ -322,14 +364,25 @@ struct _PgmImageClass {
   PgmDrawableClass parent_class;
 
   /*< private >*/
+  GThreadPool *pixbuf_loader_pool;
 
    /* Signals */
-  void (*pixbuf_loaded) (PgmImage *image);
+
+  void (*file_loaded) (PgmImage *image);
+
+  void (*cloned)      (PgmImage *image,
+                       PgmImage *clone,
+                       guint n_clones);
+
+  void (*un_cloned)   (PgmImage *image,
+                       guint n_clones);
 
   /* Padding */
   gpointer _pgm_reserved[PGM_PADDING];
 };
 
+/* Public functions */
+
 GType        pgm_image_get_type               (void);
 
 PgmDrawable *pgm_image_new                    (void);
@@ -337,6 +390,8 @@ PgmDrawable *pgm_image_new                    (void);
 PgmDrawable *pgm_image_new_from_file          (const gchar *filename,
                                                guint max_size);
 
+PgmDrawable *pgm_image_new_from_pixbuf        (GdkPixbuf *pixbuf);
+
 PgmDrawable *pgm_image_new_from_buffer        (PgmImagePixelFormat format,
                                                guint width,
                                                guint height,
@@ -350,6 +405,9 @@ PgmError     pgm_image_set_from_file          (PgmImage *image,
                                                const gchar *filename,
                                                guint max_size);
 
+PgmError     pgm_image_set_from_pixbuf        (PgmImage *image,
+                                               GdkPixbuf *pixbuf);
+
 PgmError     pgm_image_set_from_buffer        (PgmImage *image,
                                                PgmImagePixelFormat format,
                                                guint width,
@@ -374,6 +432,9 @@ PgmError     pgm_image_set_from_system_buffer (PgmImage *image,
 PgmError     pgm_image_set_from_image         (PgmImage *image,
                                                PgmImage *src_image);
 
+PgmError     pgm_image_to_pixbuf              (PgmImage *image,
+                                               GdkPixbuf **pixbuf);
+
 PgmError     pgm_image_clear                  (PgmImage *image);
 
 PgmError     pgm_image_get_storage_type       (PgmImage *image,
@@ -402,6 +463,13 @@ PgmError     pgm_image_set_interp             (PgmImage *image,
 PgmError     pgm_image_get_interp             (PgmImage *image,
                                                PgmImageInterpType *interp);
 
+PgmError     pgm_image_set_wrapping           (PgmImage *image,
+                                               PgmImageWrapping wrap_s,
+                                               PgmImageWrapping wrap_t);
+PgmError     pgm_image_get_wrapping           (PgmImage *image,
+                                               PgmImageWrapping *wrap_s,
+                                               PgmImageWrapping *wrap_t);
+
 PgmError     pgm_image_set_aspect_ratio       (PgmImage *image,
                                                guint numerator,
                                                guint denominator);
@@ -448,6 +516,11 @@ PgmError     pgm_image_to_drawable            (PgmImage *image,
                                                gint x_image,
                                                gint y_image);
 
+/* Protected functions */
+
+PgmError _pgm_image_stored_from_file_free (PgmImage *image);
+
+PgmError _pgm_image_stored_from_file_load (PgmImage *image);
 
 G_END_DECLS
 
