@@ -158,6 +158,7 @@ struct __PgmIOWatch {
  * @canvas_ratio: the attached canvas ratio with pixel-aspect-ratio correction
  * applied.
  * @caps_mask: the viewport mask of supported capacities.
+ * @message_filter: the filter of message events.
  *
  * The #PgmViewport structure.
  */
@@ -193,6 +194,8 @@ struct _PgmViewport {
 
   gulong caps_mask;
 
+  GList *message_filter;
+
   /*< private >*/
   /* Event handling */
   _PgmIOWatch *event_watch;
@@ -233,6 +236,7 @@ struct _PgmViewport {
  * @drag_motion_event: the "drag-motion-event" signal.
  * @drag_drop_event: the "drag-drop-event" signal.
  * @drag_leave_event: the "drag-leave-event" signal.
+ * @win32_message_event: the "win32-message-event" signal.
  * @pixels_read: the "pixels-read" signal.
  * @update_pass: the "update-pass" signal.
  * @set_title: the set_title virtual method.
@@ -257,6 +261,7 @@ struct _PgmViewport {
  * @get_pixel_formats: the get_pixel_formats virtual method.
  * @get_caps_mask: the get_caps_mask virtual method.
  * @get_frame_rate: the get_frame_rate virtual method.
+ * @set_message_filter: the set_message_filter virtual method.
  * @read_pixels: the read_pixels virtual method.
  * @set_drag_status: the set_drag_status virtual method.
  *
@@ -308,6 +313,9 @@ struct _PgmViewportClass {
   void     (*drag_leave_event)      (PgmViewport *viewport,
                                      PgmEventDnd *event);
 
+  void     (*win32_message_event)   (PgmViewport *viewport,
+                                     PgmEventWin32Message *event);
+
   void     (*pixels_read)           (PgmViewport *viewport,
                                      guint width,
                                      guint height,
@@ -381,6 +389,9 @@ struct _PgmViewportClass {
   PgmError (*get_frame_rate)        (PgmViewport *viewport,
                                      guint *frame_rate);
 
+  PgmError (*set_message_filter)    (PgmViewport *viewport,
+                                     GList *filter);
+
   PgmError (*read_pixels)           (PgmViewport *viewport,
                                      guint x,
                                      guint y,
@@ -477,6 +488,11 @@ PgmError     pgm_viewport_set_canvas_reflection (PgmViewport *viewport,
 PgmError     pgm_viewport_get_canvas_reflection (PgmViewport *viewport,
                                                  PgmViewportReflection *reflection);
 
+PgmError     pgm_viewport_set_message_filter    (PgmViewport *viewport,
+                                                 GList *filter);
+PgmError     pgm_viewport_get_message_filter    (PgmViewport *viewport,
+                                                 GList **filter);
+
 PgmError     pgm_viewport_update_projection     (PgmViewport *viewport);
 
 PgmError     pgm_viewport_to_canvas             (PgmViewport *viewport,
