@@ -189,6 +189,7 @@ static PgmContextProcAddress gl_proc_address = {
   (pgm_gl_bind_texture)          glBindTexture,
   (pgm_gl_tex_image_2d)          glTexImage2D,
   (pgm_gl_tex_parameter_i)       glTexParameteri,
+  (pgm_gl_tex_parameter_fv)      glTexParameterfv,
   (pgm_gl_copy_tex_sub_image_2d) glCopyTexSubImage2D,
   (pgm_gl_get_integer_v)         glGetIntegerv,
   (pgm_gl_get_float_v)           glGetFloatv,
@@ -1213,7 +1214,7 @@ render_loop (gpointer data)
   if (!pgm_backend_create_window (context->backend))
     {
       g_print ("ERROR: couldn't create the window, exiting...\n");
-      exit (1); /* FIXME: Improve error handling!!! */
+      goto error_handler;
     }
 
   context->gl = gl = &gl_proc_address;
@@ -1234,14 +1235,14 @@ render_loop (gpointer data)
   if (!context->version_string)
     {
       g_print ("ERROR: couldn't retrieve OpenGL version, exiting...\n");
-      exit (1); /* FIXME: Improve error handling!!! */
+      goto error_handler;
     }
   context->version = get_float_opengl_version (context->version_string);
   if (context->version < 1.2f)
     {
       g_print ("ERROR: OpenGL version %.1f (1.2 minimum required), exiting...\n",
                context->version);
-      exit (1); /* FIXME: Improve error handling!!! */
+      goto error_handler;
     }
 
   /* Retrieve and bind supported extensions */
@@ -1352,6 +1353,20 @@ render_loop (gpointer data)
   context->backend = NULL;
 
   return NULL;
+
+  /* FIXME: The current Pigment architecture doesn't allow to correctly report
+   *        to the plugin system that the OpenGL context can't be created. At
+   *        the moment we shamefully stop the whole process... On Windows it's
+   *        been decided to first display a pop-up explaining the issue. */
+ error_handler:
+#ifdef HAVE_WGL_H
+  MessageBox (NULL,
+              "This application cannot run on your system. The issue comes "
+              "from either your graphical card which might not be powerful "
+              "enough to support it, or from a lack of graphical driver for "
+              "it to run correctly.", NULL, MB_OK | MB_ICONERROR);
+#endif /* HAVE_WGL_H */
+  exit (1);
 }
 
 /* Initialize context creating the rendering thread */
