@@ -118,6 +118,29 @@ update_interp (PgmGlImage *glimage)
   GST_OBJECT_UNLOCK (image);
 }
 
+/* Updates the wrapping */
+static void
+update_wrapping (PgmGlImage *glimage)
+{
+  PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (glimage);
+  PgmImage *image = PGM_IMAGE (gldrawable->drawable);
+  PgmTexture *texture = glimage->texture;
+
+  GST_OBJECT_LOCK (image);
+
+  if (image->wrap_s == PGM_IMAGE_CLAMP)
+    texture->wrap_s = PGM_GL_CLAMP_TO_BORDER;
+  else if (image->wrap_s == PGM_IMAGE_REPEAT)
+    texture->wrap_s = PGM_GL_REPEAT;
+
+  if (image->wrap_t == PGM_IMAGE_CLAMP)
+    texture->wrap_t = PGM_GL_CLAMP_TO_BORDER;
+  else if (image->wrap_t == PGM_IMAGE_REPEAT)
+    texture->wrap_t = PGM_GL_REPEAT;
+
+  GST_OBJECT_UNLOCK (image);
+}
+
 /* Updates the mapping (texture) matrix */
 static void
 update_mapping_matrix (PgmGlImage *glimage)
@@ -663,6 +686,7 @@ pgm_gl_image_sync (PgmGlDrawable *gldrawable)
   /* Synchronize various properties */
   pgm_gl_image_set_fg_color (gldrawable);
   update_interp (glimage);
+  update_wrapping (glimage);
   update_alignment (glimage);
   update_last_position (glimage);
   update_drawable_ratio (glimage);
@@ -679,6 +703,9 @@ pgm_gl_image_sync (PgmGlDrawable *gldrawable)
   /* Synchronize the buffer */
   switch (type)
     {
+    case PGM_IMAGE_FILE:
+      pgm_gl_image_set_from_file (glimage);
+      break;
     case PGM_IMAGE_BUFFER:
       pgm_gl_image_set_from_buffer (glimage);
       break;
@@ -879,6 +906,49 @@ pgm_gl_image_clear (PgmGlImage *glimage)
   GST_OBJECT_UNLOCK (glimage);
 }
 
+void
+pgm_gl_image_set_from_file (PgmGlImage *glimage)
+{
+  PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (glimage);
+  PgmImage *image = PGM_IMAGE (gldrawable->drawable);
+  PgmContextTask *task;
+
+  GST_DEBUG_OBJECT (glimage, "set_from_file");
+
+  GST_OBJECT_LOCK (image);
+
+  if (G_UNLIKELY (!(image->storage_type == PGM_IMAGE_FILE
+                    || image->storage_type == PGM_IMAGE_IMAGE)))
+    {
+      GST_OBJECT_UNLOCK (image);
+      return;
+    }
+
+  if (image->data.file.pixbuf == NULL)
+    {
+      GST_OBJECT_UNLOCK (image);
+      _pgm_image_stored_from_file_load (image);
+      return;
+    }
+
+  pgm_texture_set_pixbuf (glimage->texture, image->data.file.pixbuf);
+  pgm_texture_set_matrix (glimage->texture, image->mapping_matrix);
+
+  GST_OBJECT_UNLOCK (image);
+
+  _pgm_image_stored_from_file_free (image);
+
+  glimage->empty = FALSE;
+  update_image_ratio (glimage);
+  update_layout (glimage);
+  update_slaves (glimage);
+
+  task = pgm_context_task_new (PGM_CONTEXT_GEN_TEXTURE, glimage->texture);
+  pgm_context_push_immediate_task (gldrawable->glviewport->context, task);
+  task = pgm_context_task_new (PGM_CONTEXT_UPLOAD_TEXTURE, glimage->texture);
+  pgm_context_push_deferred_task (gldrawable->glviewport->context, task);
+}
+
 void
 pgm_gl_image_set_from_buffer (PgmGlImage *glimage)
 {
@@ -1141,6 +1211,20 @@ pgm_gl_image_set_interp (PgmGlImage *glimage)
   pgm_context_push_immediate_task (gldrawable->glviewport->context, task);
 }
 
+void
+pgm_gl_image_set_wrapping (PgmGlImage *glimage)
+{
+  PgmGlDrawable *gldrawable = PGM_GL_DRAWABLE (glimage);
+  PgmContextTask *task;
+
+  GST_DEBUG_OBJECT (glimage, "set_wrapping");
+
+  update_wrapping (glimage);
+
+  task = pgm_context_task_new (PGM_CONTEXT_UPDATE_TEXTURE, glimage->texture);
+  pgm_context_push_immediate_task (gldrawable->glviewport->context, task);
+}
+
 void
 pgm_gl_image_set_aspect_ratio (PgmGlImage *glimage)
 {
