@@ -1248,6 +1248,8 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
   /* PgmGlImage changed table */
   klass->changed_func[PGM_IMAGE_DATA_EMPTY] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_clear);
+  klass->changed_func[PGM_IMAGE_DATA_FILE] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_file);
   klass->changed_func[PGM_IMAGE_DATA_BUFFER] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_from_buffer);
   klass->changed_func[PGM_IMAGE_DATA_GST_BUFFER] =
@@ -1268,6 +1270,8 @@ pgm_gl_viewport_class_init (PgmGlViewportClass *klass)
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_layout);
   klass->changed_func[PGM_IMAGE_INTERP] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_interp);
+  klass->changed_func[PGM_IMAGE_WRAPPING] =
+    GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_wrapping);
   klass->changed_func[PGM_IMAGE_ASPECT_RATIO] =
     GST_DEBUG_FUNCPTR ((PgmGlDrawableChangedFunc) pgm_gl_image_set_aspect_ratio);
   klass->changed_func[PGM_IMAGE_BORDER_WIDTH] =
