@@ -81,30 +81,48 @@ static PgmTextureUploadFunc upload_func[256] = { NULL };
 static PgmTextureBindFunc bind_func[256] = { NULL };
 static PgmTextureBindFunc unbind_func[256] = { NULL };
 
+/* Texture border colors */
+static gfloat rgb_border_color[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
+static gfloat y_border_color[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
+static gfloat cbcr_border_color[4] = { 0.5f, 0.5f, 0.5f, 0.0f };
+
 /* Keep track of the context and OpenGL procedures */
 static const PgmContextProcAddress *gl = NULL;
 static const PgmContext *context = NULL;
 
 /* Private functions  */
 
-/* Returns the upper power-of-two value of the given number */
+/* Returns number rounded up to the next power of 2. This is the branch-free
+ * implementation detailed in "Hacker's Delight" by Henry S. Warren, Jr. */
 static guint
 get_upper_power_of_two (guint number)
 {
-  guint power_of_two = 1;
-
-  while (power_of_two < number)
-    power_of_two <<= 1;
-
-  return power_of_two;
+  number = number - 1;
+  number = number | (number >> 1);
+  number = number | (number >> 2);
+  number = number | (number >> 4);
+  number = number | (number >> 8);
+  number = number | (number >> 16);
+  return number + 1;
 }
 
 /* Specifies texture parameters of the currently bound texture */
 static void
-set_texture_parameters (PgmTexture *texture)
+set_texture_parameters (PgmTexture *texture,
+                        gfloat *border_color)
 {
-  gl->tex_parameter_i (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_WRAP_S, texture->wrap);
-  gl->tex_parameter_i (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_WRAP_T, texture->wrap);
+  /* Texture border color */
+  if (border_color)
+    gl->tex_parameter_fv (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_BORDER_COLOR,
+                          border_color);
+
+  /* Texture wrapping */
+  gl->tex_parameter_i (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_WRAP_S,
+                       texture->wrap_s);
+  gl->tex_parameter_i (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_WRAP_T,
+                       texture->wrap_t);
+
+  /* Texture filtering */
   gl->tex_parameter_i (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_MIN_FILTER,
                        texture->filter);
   gl->tex_parameter_i (PGM_GL_TEXTURE_2D, PGM_GL_TEXTURE_MAG_FILTER,
@@ -123,7 +141,7 @@ do_rgb_create (PgmTexture *texture)
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
                     PGM_GL_RGBA, PGM_GL_UNSIGNED_BYTE, pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, rgb_border_color);
 
   g_free (pixels);
 }
@@ -140,7 +158,7 @@ do_bgr_create (PgmTexture *texture)
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
                     PGM_GL_BGRA, PGM_GL_UNSIGNED_BYTE, pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, rgb_border_color);
 
   g_free (pixels);
 }
@@ -155,7 +173,7 @@ do_rgba_create (PgmTexture *texture)
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
                     PGM_GL_RGBA, PGM_GL_UNSIGNED_BYTE, pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, rgb_border_color);
 
   g_free (pixels);
 }
@@ -170,7 +188,7 @@ do_bgra_create (PgmTexture *texture)
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_RGBA,
                     texture->width_pot, texture->height_pot, 0,
                     PGM_GL_BGRA, PGM_GL_UNSIGNED_BYTE, pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, rgb_border_color);
 
   g_free (pixels);
 }
@@ -207,21 +225,21 @@ do_planar_12_create (PgmTexture *texture)
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_LUMINANCE,
                     texture->width_pot, texture->height_pot, 0,
                     PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, y_pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, y_border_color);
 
   /* Cb plane creation */
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[1]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_LUMINANCE,
                     lower_width_pot, lower_height_pot, 0,
                     PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, cbcr_pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, cbcr_border_color);
 
   /* Cr plane creation */
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[2]);
   gl->tex_image_2d (PGM_GL_TEXTURE_2D, 0, PGM_GL_LUMINANCE,
                     lower_width_pot , lower_height_pot, 0,
                     PGM_GL_LUMINANCE, PGM_GL_UNSIGNED_BYTE, cbcr_pixels);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, cbcr_border_color);
 
   /* Free initialization pixels area */
   g_free (y_pixels);
@@ -475,7 +493,8 @@ init_texture (PgmTexture *texture)
   /* Parameters */
   texture->matrix = pgm_mat4x4_new_identity ();
   texture->filter = PGM_GL_LINEAR;
-  texture->wrap = PGM_GL_CLAMP_TO_EDGE;
+  texture->wrap_s = PGM_GL_REPEAT;
+  texture->wrap_t = PGM_GL_REPEAT ;
 
   /* State flags */
   texture->flags = PGM_TEXTURE_IDENTITY_MATRIX;
@@ -753,30 +772,38 @@ pgm_texture_clean (PgmTexture *texture)
 void
 pgm_texture_upload (PgmTexture *texture)
 {
-  void *buffer = NULL;
+  void *buffer;
 
-  if (!texture->id)
+  if (G_UNLIKELY (!texture->id))
     return;
 
   switch (texture->storage)
     {
     case PGM_TEXTURE_GST_BUFFER:
       buffer = GST_BUFFER_DATA (texture->data.gstbuffer);
-      upload_func[texture->csp] (texture, buffer);
       break;
 
     case PGM_TEXTURE_PIXBUF:
       buffer = gdk_pixbuf_get_pixels (texture->data.pixbuf);
-      upload_func[texture->csp] (texture, buffer);
       break;
 
     case PGM_TEXTURE_BUFFER:
-      upload_func[texture->csp] (texture, texture->data.buffer);
+      buffer = texture->data.buffer;
       break;
 
     default:
-      break;
+      return;
     }
+
+  /* FIXME: This test must not be there! At that point the content must be valid
+   * in every case. The issue is that it rarely appears that a buffer coming
+   * from a GdkPixbuf is NULL, causing a crash in the OpenGL texture update
+   * function. That is a workaround to fix this issue, the real source of the
+   * problem still need to be found though. */
+  if (G_LIKELY (buffer != NULL))
+    upload_func[texture->csp] (texture, buffer);
+
+  free_buffer (texture);
 }
 
 void
@@ -786,7 +813,7 @@ pgm_texture_update (PgmTexture *texture)
     return;
 
   gl->bind_texture (PGM_GL_TEXTURE_2D, texture->id[0]);
-  set_texture_parameters (texture);
+  set_texture_parameters (texture, NULL);
 }
 
 void
